#include <SPI.h>
#include "DRV8462.h"

#define SPI_Speed SPI_CLOCK_DIV16

uint8_t cmd = 0;
uint8_t TxData = 0;
uint16_t cmdPlusTxData = 0;

uint8_t status = 0;
uint8_t RxData = 0;
uint16_t statusPlusRxData = 0;

uint16_t result = 0;

const byte or2READ = 0b01000000;     // DRV8464's read command
const byte and2WRITE = 0b10111111;   // DRV8464's write command
const byte and2STANDARD = 0b01111111;   // DRV8464's Standard frame command

const int chipSelectPin = 53;
const int _nsleep = 8;

int8_t SPICS = 53;
int8_t SPICLK = 52;
int8_t SPIMOSI = 51;
int8_t SPIMISO = 50;

const int Steps2Move = 100000;

bool UpdateRegCurVal = false;
bool debugFlag = false;

void writeRegister(byte thisRegister, byte thisValue);
uint16_t readRegister(byte thisRegister);
void setupDRV8462(void);
void RunSingleStepTroughSPI(void);
void changeDirTroughSPI(uint8_t dir);
void RunAutoTorqueLearningRoutine(void);

void setup() {
  Serial.begin(115200);
  pinMode(chipSelectPin, OUTPUT);
  delay(5000);
  // start the SPI library:
  SPI.begin();
  SPI.setClockDivider(SPI_CLOCK_DIV2);
  SPI.setDataMode(SPI_MODE1);
  delay(5000);
  
  setupDRV8462();
  // give the sensor time to set up:
  delay(2000);
  Serial.println("DRV8462 Setup Done!");

  RunAutoTorqueLearningRoutine();
  // while(1);
}

void loop() {
  //Select High Resolution Mode
  // changeDirTroughSPI(0);
  // for(int i = 0; i < Steps2Move; i++){
    RunSingleStepTroughSPI();
    delayMicroseconds(100);
  // }

  // changeDirTroughSPI(0);
  // for(int i = 0; i < Steps2Move; i++){
  //   RunSingleStepTroughSPI();
  //   delayMicroseconds(100);
  // }
  if(debugFlag){
    Serial.println("Cycle End of" + String(Steps2Move) + "Steps!");
  }
}

void setupDRV8462(void){

  beginSPI();
  //enabling all outputs only
  // uint8_t EN_OUT_only = (DRV8xx2_REG_t.ctrl1_reg |= (1 << 7)); // bitshift by 7 to get just EN_OUT (bit 7)
  // writeRegister(SPI_CTRL1, EN_OUT_only);

  // Setting Microstep Mode
  // DRV8462_microstep_t microstep = MicroStep8;
  // uint8_t MICROSTEP_only = (DRV8xx2_REG_t.ctrl2_reg & ~MICROSTEP_MODE_MASK) | microstep;
  // writeRegister(SPI_CTRL2, MICROSTEP_only);

  //------------------------------------------------------------------------------------------------

  //Setting Auto Microstepping
  uint8_t Set_EN_AUTO_Bit = 0x01;
  uint8_t Set_EN_AUTO_Bit_only = (DRV8xx2_REG_t.ctrl1_reg & ~EN_AUTO_MASK) | Set_EN_AUTO_Bit;
  writeRegister(SPI_CTRL9, Set_EN_AUTO_Bit_only);

  //------------------------------------------------------------------------------------------------

  //Setting SPI to Control Step
  uint8_t SetSPI2CtrlStep = 0x10;
  uint8_t SPIStep_only = (DRV8xx2_REG_t.ctrl2_reg & ~SPI_STEP_MASK) | SetSPI2CtrlStep;
  writeRegister(SPI_CTRL2, SPIStep_only);

  //------------------------------------------------------------------------------------------------

  //Setting SPI to Control Dir
  uint8_t SetSPI2CtrlDir = 0x20;
  uint8_t SPIDir_only = (DRV8xx2_REG_t.ctrl2_reg & ~SPI_DIR_MASK) | SetSPI2CtrlDir;
  writeRegister(SPI_CTRL2, SPIDir_only);

  //------------------------------------------------------------------------------------------------

  // Desabling Stall Detection and report on nFault Pin
  uint8_t DiStallDetectAndReportingOnFaultPin = 0b11110111;
  uint8_t changeCTRL4 = (DRV8xx2_REG_t.ctrl4_reg & DiStallDetectAndReportingOnFaultPin);
  writeRegister(SPI_CTRL4, changeCTRL4);

  //-----------------------------------------------------------------------------------------------------------

  //Setting Output Rise and fall time
  // uint8_t SetSRBit = 0x40;
  // uint8_t SetSRBit_only = (DRV8xx2_REG_t.ctrl1_reg & ~SR_MASK) | SetSRBit;
  // writeRegister(SPI_CTRL1, SetSRBit_only);

  //------------------------------------------------------------------------------------------------

  //Setting Step Active Edge to both rising and falling edges
  // uint8_t SetStepEdgeBit = 0x08;
  // uint8_t SetStepEdgeBit_only = (DRV8xx2_REG_t.ctrl1_reg & ~(0x08)) | SetStepEdgeBit;
  // writeRegister(SPI_CTRL9, SetStepEdgeBit_only);

  //------------------------------------------------------------------------------------------------

  //Setting VREF_INT_EN bit in CTRL13 to ignore voltage on Vref Pin
  uint8_t Set_VREF_INT_EN = 0x02;
  uint8_t Set_VREF_INT_EN_only = (DRV8xx2_REG_t.ctrl13_reg & ~VREF_INT_EN_MASK) | Set_VREF_INT_EN;
  writeRegister(SPI_CTRL13, Set_VREF_INT_EN_only);

  //------------------------------------------------------------------------------------------------

  // Clearing SPIFaults
  writeRegister(SPI_CTRL3, 0xB8); // Clear fault via CTRL3 
  result = readRegister(SPI_CTRL3);
  if(debugFlag)Serial.print("result of Clearing Fault : ");Serial.print(result, HEX);Serial.print(" : ");Serial.println(result, BIN);

  //------------------------------------------------------------------------------------------------
}

bool beginSPI(void){
  int8_t _cs = SPICS;
  int8_t _clk = SPICLK;
  int8_t _miso = SPIMISO;
  int8_t _mosi = SPIMOSI;

  // Set nSLEEP to HIGH to wake up the driver
  pinMode(_nsleep, OUTPUT);  
  digitalWrite(_nsleep,  1);

  if (_clk == -1) {
    SPI.setClockDivider(SPI_CLOCK_DIV2);
    SPI.setDataMode(SPI_MODE1);
    Serial.println("Clock Devider And Mode are Set!");
  } 
  else {
    pinMode(_clk, OUTPUT);
    digitalWrite(_clk, LOW);
    pinMode(_mosi, OUTPUT);
    digitalWrite(_mosi, HIGH);
    pinMode(_miso, INPUT);
    Serial.println("Basic SPI Setting Done!");
  }
  pinMode(_cs, OUTPUT);
  digitalWrite(_cs, HIGH);
  delay(1);

  Serial.println("Reading Defaults : ======================================================================");
  readRegister(SPI_FAULT);
  readRegister(SPI_DIAG1);
  readRegister(SPI_DIAG2);
  readRegister(SPI_DIAG3);
  readRegister(SPI_CTRL2);
  readRegister(SPI_CTRL3);
  readRegister(SPI_CTRL4);
  readRegister(SPI_CTRL13);
  Serial.println("Reading Defaults Done : ======================================================================");

  // Set EN_OUT bit to 1
  uint8_t EN_OUT_only = (DRV8xx2_REG_t.ctrl1_reg |= (1 << 7)); // bitshift by 7 to get just EN_OUT (bit 7)
  writeRegister(SPI_CTRL1, EN_OUT_only);
  return true;
}

void RunSingleStepTroughSPI(void){
  //Setting Step bit in SPI_CTRL2 to run  single Step
  uint8_t SetStep = 0x40;
  uint8_t SetStep_only = (DRV8xx2_REG_t.ctrl2_reg & ~STEP_MASK) | SetStep;
  writeRegister(SPI_CTRL2, SetStep_only);
  // result = readRegister(SPI_CTRL2);
  if(statusPlusRxData & 0xFF00){
    result = readRegister(SPI_FAULT);
    result = readRegister(SPI_DIAG1);
    result = readRegister(SPI_DIAG2);
    result = readRegister(SPI_DIAG3);
  }
}

void changeDirTroughSPI(uint8_t dir){
  //Setting Step bit in SPI_CTRL2 to run  single Step
  uint8_t NewDir = dir << 7;
  uint8_t NewDir_only = (DRV8xx2_REG_t.ctrl2_reg & ~DIR_MASK) | NewDir;
  writeRegister(SPI_CTRL2, NewDir_only);
  // readRegister(SPI_CTRL2);
}

uint16_t readRegister(byte thisRegister) {
  /*
  The SDO output data word is 16 bits long and consists of the following format, as shown in Table 7-33:
  • Status byte (first 8 bits)
        – 2 MSB bits are forced high (B15, B14 = 1).
        – Following 6 bits are from the FAULT register (bits B13 through B8)
  • Report byte (last 8 bits)
        – The last 8 bits (B7:B0) are either the data currently in the register being read for a read operation (W0 =1),
          or, existing data in the register being written to for a write command (W0 = 0).
  */
  bool nameReg = true, UnNamedRegflag = false;
  String RegName = "";
  if (nameReg && debugFlag){
    if(thisRegister == SPI_FAULT){
      RegName = "SPI_FAULT";
    }
    else if(thisRegister == SPI_DIAG1){
      RegName = "SPI_DIAG1";
    }
    else if(thisRegister == SPI_DIAG2){
      RegName = "SPI_DIAG2";
    }
    else if(thisRegister == SPI_DIAG3){
      RegName = "SPI_DIAG3";
    }
    else if(thisRegister == SPI_CTRL1){
      RegName = "SPI_CTRL1";
    }
    else if(thisRegister == SPI_CTRL2){
      RegName = "SPI_CTRL2";
    }
    else if(thisRegister == SPI_CTRL3){
      RegName = "SPI_CTRL3";
    }
    else if(thisRegister == SPI_CTRL4){
      RegName = "SPI_CTRL4";
    }
    else if(thisRegister == SPI_CTRL5){
      RegName = "SPI_CTRL5";
    }
    else if(thisRegister == SPI_CTRL6){
      RegName = "SPI_CTRL6";
    }
    else if(thisRegister == SPI_CTRL7){
      RegName = "SPI_CTRL7";
    }
    else if(thisRegister == SPI_CTRL8){
      RegName = "SPI_CTRL8";
    }
    else if(thisRegister == SPI_CTRL9){
      RegName = "SPI_CTRL9";
    }
    else if(thisRegister == SPI_CTRL10){
      RegName = "SPI_CTRL10";
    }
    else if(thisRegister == SPI_CTRL11){
      RegName = "SPI_CTRL11";
    }
    else if(thisRegister == SPI_CTRL12){
      RegName = "SPI_CTRL12";
    }
    else if(thisRegister == SPI_CTRL13){
      RegName = "SPI_CTRL13";
    }
    else if(thisRegister == SPI_CTRL14){
      RegName = "SPI_CTRL14";
    }

    //Auto Torque Registers
    else if(thisRegister == SPI_ATQ_CTRL1){
      RegName = "SPI_ATQ_CTRL1";
    }
    else if(thisRegister == SPI_ATQ_CTRL2){
      RegName = "SPI_ATQ_CTRL2";
    }
    else if(thisRegister == SPI_ATQ_CTRL3){
      RegName = "SPI_ATQ_CTRL3";
    }
    else if(thisRegister == SPI_ATQ_CTRL4){
      RegName = "SPI_ATQ_CTRL4";
    }
    else if(thisRegister == SPI_ATQ_CTRL5){
      RegName = "SPI_ATQ_CTRL5";
    }
    else if(thisRegister == SPI_ATQ_CTRL6){
      RegName = "SPI_ATQ_CTRL6";
    }
    else if(thisRegister == SPI_ATQ_CTRL7){
      RegName = "SPI_ATQ_CTRL7";
    }
    else if(thisRegister == SPI_ATQ_CTRL8){
      RegName = "SPI_ATQ_CTRL8";
    }
    else if(thisRegister == SPI_ATQ_CTRL9){
      RegName = "SPI_ATQ_CTRL9";
    }
    else if(thisRegister == SPI_ATQ_CTRL10){
      RegName = "SPI_ATQ_CTRL10";
    }
    else if(thisRegister == SPI_ATQ_CTRL11){
      RegName = "SPI_ATQ_CTRL11";
    }
    else if(thisRegister == SPI_ATQ_CTRL12){
      RegName = "SPI_ATQ_CTRL12";
    }
    else if(thisRegister == SPI_ATQ_CTRL13){
      RegName = "SPI_ATQ_CTRL13";
    }
    else if(thisRegister == SPI_ATQ_CTRL14){
      RegName = "SPI_ATQ_CTRL14";
    }
    else if(thisRegister == SPI_ATQ_CTRL15){
      RegName = "SPI_ATQ_CTRL15";
    }
    else if(thisRegister == SPI_ATQ_CTRL16){
      RegName = "SPI_ATQ_CTRL16";
    }
    else if(thisRegister == SPI_ATQ_CTRL17){
      RegName = "SPI_ATQ_CTRL17";
    }
    else if(thisRegister == SPI_ATQ_CTRL18){
      RegName = "SPI_ATQ_CTRL18";
    }

    else{
      UnNamedRegflag = true;
      RegName = "Un-Named Reg";
    }
  }

  byte read_cmd = thisRegister;
  read_cmd = read_cmd & and2STANDARD; // writing 0 on MSB & MSB-1 bit to make it a standard write frame
  read_cmd = read_cmd | or2READ; // writing 0 on MSB & MSB-1 bit to make it a standard write frame
  cmdPlusTxData = ((uint16_t)read_cmd << 8); // Making LSB Byte Null data into 16 bits

  digitalWrite(chipSelectPin, LOW);
  statusPlusRxData = SPI.transfer16(cmdPlusTxData); // send the device the register you want to read:
  digitalWrite(chipSelectPin, HIGH);

  if(UpdateRegCurVal){

    if(debugFlag){
      Serial.println("Updating Register's current values After Writing Register!!!");
      Serial.print("Response from " + RegName + " : ");Serial.print(thisRegister);Serial.print(" : statusPlusRxData : ");Serial.print(statusPlusRxData, HEX);Serial.print(" : ");Serial.println(statusPlusRxData, BIN);
    }
    uint8_t RegCurrVal = (statusPlusRxData & 0x00FF);
    if(thisRegister == SPI_CTRL1){
      DRV8xx2_REG_t.ctrl1_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_CTRL2){
      DRV8xx2_REG_t.ctrl2_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_CTRL3){
      DRV8xx2_REG_t.ctrl3_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_CTRL4){
      DRV8xx2_REG_t.ctrl4_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_CTRL5){
      DRV8xx2_REG_t.ctrl5_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_CTRL6){
      DRV8xx2_REG_t.ctrl6_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_CTRL9){
      DRV8xx2_REG_t.ctrl9_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_CTRL10){
      DRV8xx2_REG_t.ctrl10_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_CTRL11){
      DRV8xx2_REG_t.ctrl11_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_CTRL12){
      DRV8xx2_REG_t.ctrl12_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_CTRL13){
      DRV8xx2_REG_t.ctrl13_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_CTRL14){
      DRV8xx2_REG_t.ctrl14_reg = RegCurrVal;
    }

    else if(thisRegister == SPI_ATQ_CTRL2){
      DRV8xx2_REG_t.atq_ctrl2_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL3){
      DRV8xx2_REG_t.atq_ctrl3_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL4){
      DRV8xx2_REG_t.atq_ctrl4_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL5){
      DRV8xx2_REG_t.atq_ctrl5_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL6){
      DRV8xx2_REG_t.atq_ctrl6_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL7){
      DRV8xx2_REG_t.atq_ctrl7_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL8){
      DRV8xx2_REG_t.atq_ctrl8_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL9){
      DRV8xx2_REG_t.atq_ctrl9_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL10){
      DRV8xx2_REG_t.atq_ctrl10_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL11){
      DRV8xx2_REG_t.atq_ctrl11_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL12){
      DRV8xx2_REG_t.atq_ctrl12_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL13){
      DRV8xx2_REG_t.atq_ctrl13_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL14){
      DRV8xx2_REG_t.atq_ctrl14_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL15){
      DRV8xx2_REG_t.atq_ctrl15_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL17){
      DRV8xx2_REG_t.atq_ctrl17_reg = RegCurrVal;
    }
    else if(thisRegister == SPI_ATQ_CTRL18){
      DRV8xx2_REG_t.atq_ctrl18_reg = RegCurrVal;
    }

    else{
      while(0);
    }

    if(debugFlag){
      if(UnNamedRegflag){
        Serial.print("Un-named register current value not updtaed! Register Address : ");Serial.println(thisRegister);
      }
      else{
        Serial.print("Current Register value updated for " + RegName + " with value : ");Serial.println(RegCurrVal, HEX);
      }
    }

    UpdateRegCurVal = false;
  }

  else{
    if(debugFlag){
      Serial.print("Response from " + RegName + " : ");Serial.print(thisRegister);Serial.print(" : statusPlusRxData : ");Serial.print(statusPlusRxData, HEX);Serial.print(" : ");Serial.println(statusPlusRxData, BIN);
    }
  }
  return (statusPlusRxData);
}

void writeRegister(byte thisRegister, byte thisValue) {
  /*
  The SDI input data word is 16 bits long and consists of the following format, as shown in Table 7-32:
    Command byte (first 8 bits)
    MSB bit indicates frame type (bit B15 = 0 for standard frame).
    Next to MSB bit, W0, indicates read or write operation (bit B14, write = 0, read = 1)
    W0 is followed by 6 address bits, A[5:0] (bits B13 through B8)
  */
  bool nameReg = true;
  String RegName = "";
  if (nameReg && debugFlag){
    if(thisRegister == SPI_FAULT){
      RegName = "SPI_FAULT";
    }
    else if(thisRegister == SPI_DIAG1){
      RegName = "SPI_DIAG1";
    }
    else if(thisRegister == SPI_DIAG2){
      RegName = "SPI_DIAG2";
    }
    else if(thisRegister == SPI_DIAG3){
      RegName = "SPI_DIAG3";
    }
    else if(thisRegister == SPI_CTRL1){
      RegName = "SPI_CTRL1";
    }
    else if(thisRegister == SPI_CTRL2){
      RegName = "SPI_CTRL2";
    }
    else if(thisRegister == SPI_CTRL3){
      RegName = "SPI_CTRL3";
    }
    else if(thisRegister == SPI_CTRL4){
      RegName = "SPI_CTRL4";
    }
    else if(thisRegister == SPI_CTRL5){
      RegName = "SPI_CTRL5";
    }
    else if(thisRegister == SPI_CTRL6){
      RegName = "SPI_CTRL6";
    }
    else if(thisRegister == SPI_CTRL7){
      RegName = "SPI_CTRL7";
    }
    else if(thisRegister == SPI_CTRL8){
      RegName = "SPI_CTRL8";
    }
    else if(thisRegister == SPI_CTRL9){
      RegName = "SPI_CTRL9";
    }
    else if(thisRegister == SPI_CTRL10){
      RegName = "SPI_CTRL10";
    }
    else if(thisRegister == SPI_CTRL11){
      RegName = "SPI_CTRL11";
    }
    else if(thisRegister == SPI_CTRL12){
      RegName = "SPI_CTRL12";
    }
    else if(thisRegister == SPI_CTRL13){
      RegName = "SPI_CTRL13";
    }
    else if(thisRegister == SPI_CTRL14){
      RegName = "SPI_CTRL14";
    }
    else{
      RegName = "Un-Named Reg";
    }
  }

  byte write_cmd = thisRegister; //register addresses will always be in lower 6 bits
  write_cmd = write_cmd & and2STANDARD; // writing 0 on MSB & MSB-1 bit to make it a standard write frame
  write_cmd = write_cmd & and2WRITE; // writing 0 on MSB & MSB-1 bit to make it a standard write frame
  cmdPlusTxData = (((uint16_t)write_cmd << 8) | thisValue); // packing data into 16 bits

  if(debugFlag){
    Serial.print("writting " + RegName + " : ");Serial.print(thisRegister);Serial.print(" : cmdPlusTxData : ");Serial.print(cmdPlusTxData, HEX);Serial.print(" : ");Serial.println(cmdPlusTxData, BIN);
  }
  
  digitalWrite(chipSelectPin, LOW); // take the chip select low to select the device:
  SPI.transfer16(cmdPlusTxData); //Send register location and data
  digitalWrite(chipSelectPin, HIGH);

  UpdateRegCurVal = true;
  readRegister(thisRegister);
  UpdateRegCurVal = false;
}

void RunAutoTorqueLearningRoutine(void){

  Serial.println("---------------------------------------Setting Up Auto Torque Parameters!!!---------------------------------------------");
  
  // //Setting TRQ_DAC register for 3A Full Scle Current
  // uint8_t Set_TRQ_DAC_3A = 0x99; // 0x99
  // uint8_t Set_TRQ_DAC = (DRV8xx2_REG_t.ctrl13_reg & (0x00)) | Set_TRQ_DAC_3A;
  // writeRegister(SPI_CTRL11, Set_TRQ_DAC);
  // // result = readRegister(SPI_CTRL11);

  //------------------------------------------------------------------------------------------------

  //Setting Auto Torque Enable
  uint8_t Set_ATQ_EN = 0x80;
  uint8_t Set_ATQ_EN_only = (DRV8xx2_REG_t.atq_ctrl10_reg & ~ATQ_EN_MASK) | Set_ATQ_EN;
  writeRegister(SPI_ATQ_CTRL10, Set_ATQ_EN_only);
  // result = readRegister(SPI_ATQ_CTRL10);

  //------------------------------------------------------------------------------------------------
  Serial.println("Running After Enabling Auto Torque");
  unsigned long ts = millis(), te = 0;
  while(true){
    if((millis() - ts) > 10000){
      break;
    }
    else{
      RunSingleStepTroughSPI();
    }
  }

  //Setting Auto Torque Initial Current = ATQ_LRN_MIN_CURRENT * 8 = 6*8 = 48 : Arount 1/3rd of Full Scale Current(3A @153)
  uint8_t Set_ATQ_LRN_MIN_CURRENT = 0x30; //0x30  // [7:3] bits of SPI_ATQ_CTRL4
  uint8_t Set_ATQ_LRN_MIN_CURRENT_only = (DRV8xx2_REG_t.atq_ctrl10_reg & ~ATQ_LRN_MIN_CURRENT_MASK) | Set_ATQ_LRN_MIN_CURRENT;
  writeRegister(SPI_ATQ_CTRL4, Set_ATQ_LRN_MIN_CURRENT_only);
  // result = readRegister(SPI_ATQ_CTRL4);

  //------------------------------------------------------------------------------------------------

  //Setting Auto Torque Final Current = ATQ_LRN_MIN_CURRENT + ATQ_LRN_STEP[1:0] = 48 + 64 : 112 < 153, Arount 3/4th of Full Scale Current(3A @153)
  uint8_t Set_ATQ_LRN_STEP = 0x0C;  // [3:2] bits of SPI_ATQ_CTRL15
  uint8_t Set_ATQ_LRN_STEP_only = (DRV8xx2_REG_t.atq_ctrl10_reg & ~ATQ_LRN_STEP_MASK) | Set_ATQ_LRN_STEP;
  writeRegister(SPI_ATQ_CTRL15, Set_ATQ_LRN_STEP_only);
  // result = readRegister(SPI_ATQ_CTRL15);

  //------------------------------------------------------------------------------------------------

  //Setting Auto Torque Learn Cycle = ATQ_LRN_CYCLE_SELECT[1:0] = 0b11, 32 Half Cycle for stable learning
  uint8_t Set_ATQ_LRN_CYCLE_SELECT = 0x03;  // [1:0] bits of SPI_ATQ_CTRL15
  uint8_t Set_ATQ_LRN_CYCLE_SELECT_only = (DRV8xx2_REG_t.atq_ctrl10_reg & ~ATQ_LRN_STEP_MASK) | Set_ATQ_LRN_CYCLE_SELECT;
  writeRegister(SPI_ATQ_CTRL15, Set_ATQ_LRN_CYCLE_SELECT_only);
  // result = readRegister(SPI_ATQ_CTRL15);

  //------------------------------------------------------------------------------------------------

  //Setting Start Auto ATQ_FRZ for most stable response
  uint8_t Set_ATQ_FRZ = 0x38;
  uint8_t Set_ATQ_FRZ_only = (DRV8xx2_REG_t.atq_ctrl10_reg & ~ATQ_FRZ_MASK) | Set_ATQ_FRZ;
  writeRegister(SPI_ATQ_CTRL10, Set_ATQ_FRZ_only);
  // result = readRegister(SPI_ATQ_CTRL10);

  //------------------------------------------------------------------------------------------------

  //Setting Start Auto ATQ_AVZ for most stable response
  uint8_t Set_ATQ_AVZ = 0x07;
  uint8_t Set_ATQ_AVZ_only = (DRV8xx2_REG_t.atq_ctrl10_reg & ~ATQ_AVG_MASK) | Set_ATQ_AVZ;
  writeRegister(SPI_ATQ_CTRL10, Set_ATQ_AVZ_only);
  // result = readRegister(SPI_ATQ_CTRL10);

  //------------------------------------------------------------------------------------------------

  //Setting Start Auto Torque Learning
  uint8_t Set_LRN_START = 0x40;
  uint8_t Set_LRN_START_MASK_only = (DRV8xx2_REG_t.atq_ctrl10_reg & ~LRN_START_MASK) | Set_LRN_START;
  writeRegister(SPI_ATQ_CTRL10, Set_LRN_START_MASK_only);
  // result = readRegister(SPI_ATQ_CTRL10);

  //------------------------------------------------------------------------------------------------

  delay(10000);
  
  Serial.println("---------------------------------------Running Auto Torque learning Routine!!!---------------------------------------------");

  uint16_t ATQ_LRN_CONST1_Val = 0, ATQ_LRN_CONST2_Val = 0, ATQ_TRQ_MAX_Val = 0, ATQ_TRQ_MIN_Val = 0;
  bool ATQ_LRN_START_Bit = 1, ATQ_LRN_DONE_Bit = 0;

  //Reading ATQ_LRN_CONST1
  result = readRegister(SPI_ATQ_CTRL3);
  ATQ_LRN_CONST1_Val = (result & 0x00FF);
  result = readRegister(SPI_ATQ_CTRL2);
  ATQ_LRN_CONST1_Val = ATQ_LRN_CONST1_Val | ((result & 0x0007) << 8);  // Clubing 9th, 10th & 11th bit of ATQ_LRN_CONST1
  if(debugFlag){
    Serial.print("ATQ_LRN_CONST1_Val : ");Serial.print(ATQ_LRN_CONST1_Val, HEX);Serial.print(" : ");Serial.println(ATQ_LRN_CONST1_Val, BIN);
  }

  //Reading ATQ_LRN_CONST2
  result = readRegister(SPI_ATQ_CTRL5);
  ATQ_LRN_CONST2_Val = (result & 0x00FF);
  result = readRegister(SPI_ATQ_CTRL4);
  ATQ_LRN_CONST2_Val = ATQ_LRN_CONST2_Val | ((result & 0x0007) << 8);  // Clubing 9th, 10th & 11th bit of ATQ_LRN_CONST2
  if(debugFlag){
    Serial.print("ATQ_LRN_CONST2_Val : ");Serial.print(ATQ_LRN_CONST2_Val, HEX);Serial.print(" : ");Serial.println(ATQ_LRN_CONST2_Val, BIN);
  }

  //Reading ATQ_TRQ_MIN
  result = readRegister(SPI_ATQ_CTRL11);
  ATQ_TRQ_MIN_Val = (result & 0x00FF);
  if(debugFlag){
    Serial.print("ATQ_TRQ_MIN_Val : ");Serial.print(ATQ_TRQ_MIN_Val, HEX);Serial.print(" : ");Serial.println(ATQ_TRQ_MIN_Val, BIN);
  }
  //Reading ATQ_TRQ_MAX
  result = readRegister(SPI_ATQ_CTRL12);
  ATQ_TRQ_MAX_Val = (result & 0x00FF);
  if(debugFlag){
    Serial.print("ATQ_TRQ_MAX_Val : ");Serial.print(ATQ_TRQ_MAX_Val, HEX);Serial.print(" : ");Serial.println(ATQ_TRQ_MAX_Val, BIN);
  }

  while(true){
    result = readRegister(SPI_DIAG2);
    ATQ_LRN_DONE_Bit = ((result >> 2) & 0x0001); //Extracting 2nd bit of DIAG2 Reg
    result = readRegister(SPI_ATQ_CTRL10);
    ATQ_LRN_START_Bit = ((result >> 6) & 0x0001); //Extracting 6th bit of ATQ_CTRL10 Reg

    if(ATQ_LRN_DONE_Bit == 1 && ATQ_LRN_START_Bit == 0){
      Serial.println("---------------------------------------Auto Torque learning Done!!!---------------------------------------------");
      //Reading ATQ_LRN_CONST1
      result = readRegister(SPI_ATQ_CTRL3);
      ATQ_LRN_CONST1_Val = (result & 0x00FF);
      result = readRegister(SPI_ATQ_CTRL2);
      ATQ_LRN_CONST1_Val = ATQ_LRN_CONST1_Val | ((result & 0x0007) << 8);  // Clubing 9th, 10th & 11th bit of ATQ_LRN_CONST1
      if(debugFlag){
        Serial.print("ATQ_LRN_CONST1_Val : ");Serial.print(ATQ_LRN_CONST1_Val, HEX);Serial.print(" : ");Serial.println(ATQ_LRN_CONST1_Val, BIN);
      }
      //Reading ATQ_LRN_CONST2
      result = readRegister(SPI_ATQ_CTRL5);
      ATQ_LRN_CONST2_Val = (result & 0x00FF);
      result = readRegister(SPI_ATQ_CTRL4);
      ATQ_LRN_CONST2_Val = ATQ_LRN_CONST2_Val | ((result & 0x0007) << 8);  // Clubing 9th, 10th & 11th bit of ATQ_LRN_CONST2
      if(debugFlag){
        Serial.print("ATQ_LRN_CONST2_Val : ");Serial.print(ATQ_LRN_CONST2_Val, HEX);Serial.print(" : ");Serial.println(ATQ_LRN_CONST2_Val, BIN);
      }

      //Reading ATQ_TRQ_MIN
      result = readRegister(SPI_ATQ_CTRL11);
      ATQ_TRQ_MIN_Val = (result & 0x00FF);
      if(debugFlag){
        Serial.print("ATQ_TRQ_MIN_Val : ");Serial.print(ATQ_TRQ_MIN_Val, HEX);Serial.print(" : ");Serial.println(ATQ_TRQ_MIN_Val, BIN);
      }
      //Reading ATQ_TRQ_MAX
      result = readRegister(SPI_ATQ_CTRL12);
      ATQ_TRQ_MAX_Val = (result & 0x00FF);
      if(debugFlag){
        Serial.print("ATQ_TRQ_MAX_Val : ");Serial.print(ATQ_TRQ_MAX_Val, HEX);Serial.print(" : ");Serial.println(ATQ_TRQ_MAX_Val, BIN);
      }
      break;
    }
    else{
      RunSingleStepTroughSPI();
      if(debugFlag){
        Serial.print("Moved for Auto Torque learning : ");
      }
      delay(8);
    }
  }
}
